/*
 * Copyright notice from original mutt:
 * Copyright (C) 1996-2002 Michael R. Elkins <me@mutt.org>
 *
 * Parts were written/modified by:
 * Rocco Rutte <pdmef@cs.tu-berlin.de>
 *
 * This file is part of mutt-ng, see http://www.muttng.org/.
 * It's licensed under the GNU General Public License,
 * please see the file GPL in the top level source directory.
 */

#include <lib-lib/lib-lib.h>

#include <lib-lua/lib-lua.h>
#include <lib-sys/unix.h>
#include <lib-ui/lib-ui.h>
#include <lib-ui/history.h>
#include <lib-mx/mx.h>

#include "mutt.h"
#include "keymap.h"
#include "crypt.h"
#include "charset.h"
#include "thread.h"
#include "mutt_idna.h"
#include "alias.h"
#include "init.h"

/*
 * prototypes
 */
static const struct mapping_t* get_sortmap (struct option_t* option);
static int parse_sort (struct option_t* dst, const char *s,
                       const struct mapping_t *map,
                       char* errbuf, ssize_t errlen);

static hash_t *ConfigOptions = NULL;

/* for synonym warning reports: synonym found during parsing */
typedef struct syn_t {
  struct syn_t *next;
  char* f;              /* file */
  int l;                /* line */
  struct option_t* n;   /* new */
  struct option_t* o;   /* old */
} syn_t;

DO_INIT(syn_t, syn);
static void syn_wipe(syn_t *syn) {
    p_delete(&syn->f);
}
DO_NEW(syn_t, syn);
DO_DELETE(syn_t, syn);
DO_SLIST(syn_t, syn, syn_delete);

/* for synonym warning reports: list of synonyms found */
static syn_t *Synonyms = NULL;
/* for synonym warning reports: current rc file */
static const char* CurRCFile = NULL;
/* for synonym warning reports: current rc line */
static int CurRCLine = 0;

/* prototypes for checking for special vars */
static int check_history    (const char* option, unsigned long val,
                             char* errbuf, ssize_t errlen);
/* this checks that numbers are >= 0 */
static int check_num        (const char* option, unsigned long val,
                             char* errbuf, ssize_t errlen);

/* use this to check only */
static int check_special (const char* option, unsigned long val,
                          char* errbuf, ssize_t errlen);

/* variable <-> sanity check function mappings
 * when changing these, make sure the proper _from_string handler
 * does this checking!
 */
static struct {
  const char* name;
  int (*check) (const char* option, unsigned long val,
                char* errbuf, ssize_t errlen);
} SpecialVars[] = {
  { "history",                  check_history },
  { "pager_index_lines",        check_num },
  /* last */
  { NULL,         NULL }
};

static void bool_to_string (char* dst, ssize_t dstlen,
                            struct option_t* option) {
  snprintf (dst, dstlen, "%s=%s", option->option,
            option (option->data) ? "yes" : "no");
}

static int bool_from_string (struct option_t* dst, const char* val,
                             char* errbuf __attribute__ ((unused)),
                             ssize_t errlen __attribute__ ((unused))) {
  int flag = -1;

  if (!dst)
    return 0;
  if (ascii_strncasecmp (val, "yes", 3) == 0)
    flag = 1;
  else if (ascii_strncasecmp (val, "no", 2) == 0)
    flag = 0;

  if (flag < 0)
    return 0;
  if (flag)
    set_option (dst->data);
  else
    unset_option (dst->data);
  return 1;
}

static void num_to_string (char* dst, ssize_t dstlen,
                           struct option_t* option) {
  /* XXX puke */
  const char* fmt = (m_strcmp(option->option, "umask") == 0) ?
                    "%s=%04o" : "%s=%d";
  snprintf (dst, dstlen, fmt, option->option,
            *((short*) option->data));
}

static int num_from_string (struct option_t* dst, const char* val,
                            char* errbuf, ssize_t errlen) {
  int num = 0, old = 0;
  char* t = NULL;

  if (!dst)
    return 0;

  num = strtol (val, &t, 0);

  if (m_strisempty(val) || *t || (short) num != num) {
    if (errbuf) {
      snprintf (errbuf, errlen, _("'%s' is invalid for $%s"),
                val, dst->option);
    }
    return 0;
  }

  /* just temporarily accept new val so that check_special for
   * $history already has it when doing history's init() */
  old = *((short*) dst->data);
  *((short*) dst->data) = (short) num;

  if (!check_special (dst->option, (unsigned long) num, errbuf, errlen)) {
    *((short*) dst->data) = old;
    return 0;
  }

  return 1;
}

static void str_to_string (char* dst, ssize_t dstlen,
                           struct option_t* option) {
  snprintf (dst, dstlen, "%s=\"%s\"", option->option,
            NONULL (*((char**) option->data)));
}

static int path_from_string (struct option_t* dst, const char* val,
                             char* errbuf __attribute__ ((unused)), ssize_t errlen __attribute__ ((unused))) {
  char path[_POSIX_PATH_MAX];

  if (!dst)
    return 0;

  if (m_strisempty(val)) {
    p_delete((char**) dst->data);
    return 1;
  }

  path[0] = '\0';
  m_strcpy(path, sizeof(path), val);
  mutt_expand_path (path, sizeof(path));
  m_strreplace((char **) dst->data, path);
  return 1;
}

static int str_from_string (struct option_t* dst, const char* val,
                            char* errbuf, ssize_t errlen) {
  if (!dst)
    return 0;

  if (!check_special (dst->option, (unsigned long) val, errbuf, errlen))
    return 0;

  m_strreplace((char**) dst->data, val);
  return 1;
}

static void quad_to_string (char* dst, ssize_t dstlen,
                            struct option_t* option) {
  const char *vals[] = { "no", "yes", "ask-no", "ask-yes" };
  snprintf (dst, dstlen, "%s=%s", option->option,
            vals[quadoption (option->data)]);
}

static int quad_from_string (struct option_t* dst, const char* val,
                             char* errbuf __attribute__ ((unused)), ssize_t errlen __attribute__ ((unused))) {
  int flag = -1;

  if (!dst)
    return 0;
  if (ascii_strncasecmp (val, "yes", 3) == 0)
    flag = M_YES;
  else if (ascii_strncasecmp (val, "no", 2) == 0)
    flag = M_NO;
  else if (ascii_strncasecmp (val, "ask-yes", 7) == 0)
    flag = M_ASKYES;
  else if (ascii_strncasecmp (val, "ask-no", 6) == 0)
    flag = M_ASKNO;

  if (flag < 0)
    return 0;

  set_quadoption (dst->data, flag);
  return 1;
}

static void sort_to_string (char* dst, ssize_t dstlen,
                            struct option_t* option) {
  const struct mapping_t *map = get_sortmap (option);
  const char *p = NULL;

  if (!map) {
    snprintf (dst, sizeof(dst), "%s=unknown", option->option);
    return;
  }

  p = mutt_getnamebyvalue(*((short *)option->data) & SORT_MASK, map);

  snprintf (dst, dstlen, "%s=%s%s%s", option->option,
            (*((short *) option->data) & SORT_REVERSE) ?
            "reverse-" : "",
            (*((short *) option->data) & SORT_LAST) ? "last-" :
            "", NONULL (p));
}

static int sort_from_string (struct option_t* dst, const char* val,
                             char* errbuf, ssize_t errlen) {
  const struct mapping_t *map = NULL;
  if (!(map = get_sortmap (dst))) {
    if (errbuf)
      snprintf (errbuf, errlen, _("%s: Unknown type."),
                dst->option);
    return 0;
  }
  if (parse_sort (dst, val, map, errbuf, errlen) == -1)
    return 0;
  return 1;
}

static void rx_to_string (char* dst, ssize_t dstlen,
                          struct option_t* option) {
  rx_t* p = (rx_t*) option->data;
  snprintf (dst, dstlen, "%s=\"%s\"", option->option,
            NONULL (p->pattern));
}

static int rx_from_string (struct option_t* dst, const char* val,
                           char* errbuf, ssize_t errlen) {
  rx_t* p = NULL;
  regex_t* rx = NULL;
  int flags = 0, e = 0, neg = 0;
  char* s = NULL;

  if (!dst)
    return 0;

  if (option (OPTATTACHMSG) && !m_strcmp(dst->option, "reply_regexp")) {
    if (errbuf)
      snprintf (errbuf, errlen,
                "Operation not permitted when in attach-message mode.");
    return 0;
  }

  if (!((rx_t*) dst->data))
    *((rx_t**) dst->data) = p_new(rx_t, 1);

  p = (rx_t*) dst->data;

  /* something to do? */
  if (m_strisempty(val) || (p->pattern && m_strcmp(p->pattern, val) == 0))
    return 1;

  if (m_strcmp(dst->option, "mask") != 0)
    flags |= mutt_which_case (val);

  s = (char*) val;
  if (m_strcmp(dst->option, "mask") == 0 && *s == '!') {
    neg = 1;
    s++;
  }

  rx = p_new(regex_t, 1);

  if ((e = REGCOMP (rx, s, flags)) != 0) {
    regerror (e, rx, errbuf, errlen);
    regfree (rx);
    p_delete(&rx);
    return 0;
  }

  if (p->rx) {
    regfree (p->rx);
    p_delete(&p->rx);
  }

  m_strreplace(&p->pattern, val);
  p->rx = rx;
  p->neg = neg;

  if (m_strcmp(dst->option, "reply_regexp") == 0)
    mutt_adjust_all_subjects ();

  return 1;
}

static int magic_from_string (struct option_t* dst, const char* val,
                              char *errbuf, ssize_t errlen)
{
  int flag = -1;

  if (!dst || m_strisempty(val))
    return 0;
  if (ascii_strncasecmp (val, "mbox", 4) == 0)
    flag = M_MBOX;
  else if (ascii_strncasecmp (val, "mh", 2) == 0)
    flag = M_MH;
  else if (ascii_strncasecmp (val, "maildir", 7) == 0)
    flag = M_MAILDIR;

  if (flag < 0)
    return 0;

  *((short*) dst->data) = flag;
  return 1;

}

static struct {
  unsigned short type;
  void (*opt_tostr) (char* dst, ssize_t dstlen, struct option_t* option);
  int (*opt_fromstr) (struct option_t* dst, const char* val,
                          char* errbuf, ssize_t errlen);
} FuncTable[] = {
  { 0,          NULL,             NULL }, /* there's no DT_ type with 0 */
  { DT_BOOL,    bool_to_string,   bool_from_string },
  { DT_NUM,     num_to_string,    num_from_string },
  { DT_STR,     str_to_string,    str_from_string },
  { DT_PATH,    str_to_string,    path_from_string },
  { DT_QUAD,    quad_to_string,   quad_from_string },
  { DT_SORT,    sort_to_string,   sort_from_string },
  { DT_RX,      rx_to_string,     rx_from_string },
};


int mutt_option_value (const char* val, char* dst, ssize_t dstlen) {
  struct option_t* option = NULL;
  char* tmp = NULL, *t = NULL;
  ssize_t l = 0;

  if (!(option = hash_find (ConfigOptions, val))) {
    *dst = '\0';
    return 0;
  }
  tmp = p_new(char, dstlen+1);
  FuncTable[DTYPE(option->type)].opt_tostr (tmp, dstlen, option);

  /* as we get things of type $var=value and don't want to bloat the
   * above "just" for expansion, we do the stripping here */
  t = strchr (tmp, '=');
  t++;
  l = m_strlen(t);
  if (l >= 2) {
    if (t[l-1] == '"' && *t == '"') {
      t[l-1] = '\0';
      t++;
    }
  }
  memcpy (dst, t, l+1);
  p_delete(&tmp);

  return 1;
}

static void toggle_quadoption (int opt)
{
  int n = opt / 4;
  int b = (opt % 4) * 2;

  QuadOptions[n] ^= (1 << b);
}

void set_quadoption (int opt, int flag)
{
  int n = opt / 4;
  int b = (opt % 4) * 2;

  QuadOptions[n] &= ~(0x3 << b);
  QuadOptions[n] |= (flag & 0x3) << b;
}

int quadoption (int opt)
{
  int n = opt / 4;
  int b = (opt % 4) * 2;

  return (QuadOptions[n] >> b) & 0x3;
}

int query_quadoption2(int v, const char *prompt)
{
  switch (v) {
  case M_YES:
  case M_NO:
    return v;

  default:
    return mutt_yesorno(prompt, (v == M_ASKYES));
  }
}

int query_quadoption (int opt, const char *prompt)
{
  int v = quadoption (opt);

  switch (v) {
  case M_YES:
  case M_NO:
    return v;

  default:
    return mutt_yesorno(prompt, (v == M_ASKYES));
  }
}

/* always wise to do what someone else did before */
static void _attachments_clean (void) {
  int i;
  if (Context && Context->msgcount) {
    for (i = 0; i < Context->msgcount; i++)
      Context->hdrs[i]->attach_valid = 0;
  }
}

static int parse_attach_list (BUFFER *buf, BUFFER *s, string_list_t **ldata,
                              BUFFER *err __attribute__ ((unused))) {
  ATTACH_MATCH *a;
  string_list_t *listp, *lastp;
  char *p;
  char *tmpminor;
  int len;

  /* Find the last item in the list that data points to. */
  lastp = NULL;
  for (listp = *ldata; listp; listp = listp->next) {
    a = (ATTACH_MATCH *)listp->data;
    lastp = listp;
  }

  do {
    mutt_extract_token (buf, s, 0);

    if (!buf->data || *buf->data == '\0')
      continue;

    a = p_new(ATTACH_MATCH, 1);

    /* some cheap hacks that I expect to remove */
    if (!m_strcasecmp(buf->data, "any"))
      a->major = m_strdup("*/.*");
    else if (!m_strcasecmp(buf->data, "none"))
      a->major = m_strdup("cheap_hack/this_should_never_match");
    else
      a->major = m_strdup(buf->data);

    if ((p = strchr(a->major, '/'))) {
      *p = '\0';
      ++p;
      a->minor = p;
    } else {
      a->minor = "unknown";
    }

    len = m_strlen(a->minor);
    tmpminor = p_new(char, len + 3);
    m_strcpy(&tmpminor[1], len + 3, a->minor);
    tmpminor[0] = '^';
    tmpminor[len+1] = '$';
    tmpminor[len+2] = '\0';

    a->major_int = mutt_check_mime_type(a->major);
    regcomp(&a->minor_rx, tmpminor, REG_ICASE|REG_EXTENDED);

    p_delete(&tmpminor);

    listp = p_new(string_list_t, 1);
    listp->data = (char *)a;
    listp->next = NULL;
    if (lastp) {
      lastp->next = listp;
    } else {
      *ldata = listp;
    }
    lastp = listp;
  }
  while (MoreArgs (s));

  _attachments_clean();
  return 0;
}

static int parse_unattach_list (BUFFER *buf, BUFFER *s, string_list_t **ldata,
                                BUFFER *err __attribute__ ((unused))) {
  ATTACH_MATCH *a;
  string_list_t *lp, *lastp, *newlp;
  char *tmp;
  int major;
  char *minor;

  do {
    mutt_extract_token (buf, s, 0);

    if (!m_strcasecmp(buf->data, "any"))
      tmp = m_strdup("*/.*");
    else if (!m_strcasecmp(buf->data, "none"))
      tmp = m_strdup("cheap_hack/this_should_never_match");
    else
      tmp = m_strdup(buf->data);

    if ((minor = strchr(tmp, '/'))) {
      *minor = '\0';
      ++minor;
    } else {
      minor = m_strdup("unknown");
    }
    major = mutt_check_mime_type(tmp);

    /* We must do our own walk here because string_list_remove() will only
     * remove the string_list_t->data, not anything pointed to by the string_list_t->data. */
    lastp = NULL;
    for(lp = *ldata; lp; ) {
      a = (ATTACH_MATCH *)lp->data;
      if (a->major_int == major && !m_strcasecmp(minor, a->minor)) {
        regfree(&a->minor_rx);
        p_delete(&a->major);

        /* Relink backward */
        if (lastp)
          lastp->next = lp->next;
        else
          *ldata = lp->next;

        newlp = lp->next;
        p_delete(&lp->data); /* same as a */
        p_delete(&lp);
        lp = newlp;
        continue;
      }

      lastp = lp;
      lp = lp->next;
    }
  }
  while (MoreArgs (s));

  p_delete(&tmp);
  _attachments_clean();
  return 0;
}

static int print_attach_list (string_list_t *lp, char op, const char *name) {
  while (lp) {
    printf("attachments %c%s %s/%s\n", op, name,
           ((ATTACH_MATCH *)lp->data)->major,
           ((ATTACH_MATCH *)lp->data)->minor);
    lp = lp->next;
  }

  return 0;
}

static int parse_attachments (BUFFER *buf, BUFFER *s,
                              unsigned long data __attribute__ ((unused)),
                              BUFFER *err) {
  char op, *category;
  string_list_t **listp;

  mutt_extract_token(buf, s, 0);
  if (!buf->data || *buf->data == '\0') {
    m_strcpy(err->data, err->dsize, _("attachments: no disposition"));
    return -1;
  }

  category = buf->data;
  op = *category++;

  if (op == '?') {
    mutt_endwin (NULL);
    fflush (stdout);
    printf("\nCurrent attachments settings:\n\n");
    print_attach_list(AttachAllow, '+', "A");
    print_attach_list(AttachExclude, '-', "A");
    print_attach_list(InlineAllow, '+', "I");
    print_attach_list(InlineExclude, '-', "I");
    set_option (OPTFORCEREDRAWINDEX);
    set_option (OPTFORCEREDRAWPAGER);
    mutt_any_key_to_continue (NULL);
    return 0;
  }

  if (op != '+' && op != '-') {
    op = '+';
    category--;
  }
  if (!m_strncasecmp(category, "attachment", strlen(category))) {
    if (op == '+')
      listp = &AttachAllow;
    else
      listp = &AttachExclude;
  }
  else if (!m_strncasecmp(category, "inline", strlen(category))) {
    if (op == '+')
      listp = &InlineAllow;
    else
      listp = &InlineExclude;
  } else {
    m_strcpy(err->data, err->dsize, _("attachments: invalid disposition"));
    return -1;
  }

  return parse_attach_list(buf, s, listp, err);
}

static int parse_unattachments (BUFFER *buf, BUFFER *s, unsigned long data __attribute__ ((unused)), BUFFER *err) {
  char op, *p;
  string_list_t **listp;

  mutt_extract_token(buf, s, 0);
  if (!buf->data || *buf->data == '\0') {
    m_strcpy(err->data, err->dsize, _("unattachments: no disposition"));
    return -1;
  }

  p = buf->data;
  op = *p++;
  if (op != '+' && op != '-') {
    op = '+';
    p--;
  }
  if (!m_strncasecmp(p, "attachment", strlen(p))) {
    if (op == '+')
      listp = &AttachAllow;
    else
      listp = &AttachExclude;
  }
  else if (!m_strncasecmp(p, "inline", strlen(p))) {
    if (op == '+')
      listp = &InlineAllow;
    else
      listp = &InlineExclude;
  }
  else {
    m_strcpy(err->data, err->dsize, _("unattachments: invalid disposition"));
    return -1;
  }

  return parse_unattach_list(buf, s, listp, err);
}

static int parse_unalias (BUFFER * buf, BUFFER * s,
                          unsigned long data __attribute__ ((unused)),
                          BUFFER * err __attribute__ ((unused)))
{
    alias_t *tmp, **last;

    do {
        mutt_extract_token (buf, s, 0);

        if (!m_strcmp("*", buf->data) == 0) {
            if (CurrentMenu == MENU_ALIAS) {
                for (tmp = Aliases; tmp; tmp = tmp->next)
                    tmp->del = 1;
                set_option(OPTFORCEREDRAWINDEX);
            } else {
                alias_list_wipe(&Aliases);
            }
            break;
        }

        last = &Aliases;
        for (last = &Aliases; *last; last = &(*last)->next) {
            if (!m_strcasecmp(buf->data, (*last)->name)) {
                if (CurrentMenu == MENU_ALIAS) {
                    (*last)->del = 1;
                    set_option (OPTFORCEREDRAWINDEX);
                } else {
                    tmp = alias_list_pop(last);
                    alias_delete(&tmp);
                }
                break;
            }
        }
    } while (MoreArgs(s));

    return 0;
}

static int parse_alias (BUFFER * buf, BUFFER * s,
                        unsigned long data __attribute__ ((unused)),
                        BUFFER * err)
{
    alias_t **last;
    char *estr = NULL;

    if (!MoreArgs (s)) {
        m_strcpy(err->data, err->dsize, _("alias: no address"));
        return -1;
    }

    mutt_extract_token (buf, s, 0);

    /* check to see if an alias with this name already exists */
    for (last = &Aliases; *last; last = &(*last)->next) {
        if (!m_strcasecmp((*last)->name, buf->data))
            break;
    }

    if (!*last) {
        /* create a new alias */
        *last = alias_new();
        (*last)->name = m_strdup(buf->data);
        /* give the main addressbook code a chance */
        if (CurrentMenu == MENU_ALIAS)
            set_option (OPTMENUCALLER);
    } else {
        /* override the previous value */
        address_list_wipe(&(*last)->addr);
        if (CurrentMenu == MENU_ALIAS)
            set_option (OPTFORCEREDRAWINDEX);
    }

    mutt_extract_token(buf, s, M_TOKEN_QUOTE | M_TOKEN_SPACE);
    (*last)->addr = mutt_parse_adrlist((*last)->addr, buf->data);
    if (mutt_addrlist_to_idna((*last)->addr, &estr)) {
        snprintf (err->data, err->dsize,
                  _("Warning: Bad IDN '%s' in alias '%s'.\n"), estr, (*last)->name);
        p_delete(&estr);
        return -1;
    }

    return 0;
}

static int
parse_unmy_hdr(BUFFER * buf, BUFFER * s,
               unsigned long data __attribute__ ((unused)),
               BUFFER * err __attribute__ ((unused)))
{
    do {
        mutt_extract_token (buf, s, 0);

        if (!m_strcmp("*", buf->data)) {
            string_list_wipe(&UserHeader);
        } else {
            string_list_t **last = &UserHeader;
            ssize_t l = m_strlen(buf->data);

            if (buf->data[l - 1] == ':')
                l--;

            while (*last) {
                if (!ascii_strncasecmp(buf->data, (*last)->data, l)
                && (*last)->data[l] == ':')
                {
                    string_list_t *tmp = string_list_pop(last);
                    string_item_delete(&tmp);
                } else {
                    last = &(*last)->next;
                }
            }
        }
    } while (MoreArgs(s));

    return 0;
}

static int parse_my_hdr (BUFFER * buf, BUFFER * s, unsigned long data __attribute__ ((unused)),
                         BUFFER * err)
{
  string_list_t *tmp;
  ssize_t keylen;
  char *p;

  mutt_extract_token (buf, s, M_TOKEN_SPACE | M_TOKEN_QUOTE);
  if ((p = strpbrk (buf->data, ": \t")) == NULL || *p != ':') {
    m_strcpy(err->data, err->dsize, _("invalid header field"));
    return -1;
  }
  keylen = p - buf->data + 1;

  if (UserHeader) {
    for (tmp = UserHeader;; tmp = tmp->next) {
      /* see if there is already a field by this name */
      if (ascii_strncasecmp (buf->data, tmp->data, keylen) == 0) {
        /* replace the old value */
        p_delete(&tmp->data);
        tmp->data = buf->data;
        p_clear(buf, 1);
        return 0;
      }
      if (!tmp->next)
        break;
    }
    tmp->next = string_item_new();
    tmp = tmp->next;
  }
  else {
    tmp = string_item_new();
    UserHeader = tmp;
  }
  tmp->data = buf->data;
  p_clear(buf, 1);
  return 0;
}

static int
parse_sort (struct option_t* dst, const char *s, const struct mapping_t *map,
            char* errbuf, ssize_t errlen) {
  int i, flags = 0;

  if (m_strncmp("reverse-", s, 8) == 0) {
    s += 8;
    flags = SORT_REVERSE;
  }

  if (m_strncmp("last-", s, 5) == 0) {
    s += 5;
    flags |= SORT_LAST;
  }

  if ((i = mutt_getvaluebyname (s, map)) == -1) {
    if (errbuf)
      snprintf (errbuf, errlen, _("'%s' is invalid for $%s"), s, dst->option);
    return -1;
  }

  *((short*) dst->data) = i | flags;
  return 0;
}

/* if additional data more == 1, we want to resolve synonyms */
static void mutt_set_default(const char *name __attribute__ ((unused)), void* p, unsigned long more)
{
    char buf[LONG_STRING];
    struct option_t *ptr = p;

    if (!ptr || *ptr->init || !FuncTable[DTYPE (ptr->type)].opt_fromstr)
        return;

    mutt_option_value(ptr->option, buf, sizeof(buf));
    if (m_strlen(ptr->init) == 0 && *buf)
        ptr->init = m_strdup(buf);
}

static int init_expand (char** dst, struct option_t* src) {
  BUFFER token, in;
  ssize_t len = 0;

  p_delete(dst);

  if (DTYPE(src->type) == DT_STR || DTYPE(src->type) == DT_PATH) {
    /* only expand for string as it's the only place where
     * we want to expand vars right now */
    if (src->init && *src->init) {
      p_clear(&token, 1);
      p_clear(&in, 1);
      len = m_strlen(src->init) + 2;
      in.data = p_new(char, len + 1);
      snprintf (in.data, len, "\"%s\"", src->init);
      in.dptr = in.data;
      in.dsize = len;
      mutt_extract_token (&token, &in, 0);
      if (token.data && *token.data)
        *dst = m_strdup(token.data);
      else
        *dst = m_strdup("");
      p_delete(&in.data);
      p_delete(&token.data);
    } else
      *dst = m_strdup("");
  } else
    /* for non-string: take value as is */
    *dst = m_strdup(src->init);
  return 1;
}

/* if additional data more == 1, we want to resolve synonyms */
static void mutt_restore_default (const char* name __attribute__ ((unused)),
                                  void* p, unsigned long more) {
  char errbuf[STRING];
  struct option_t* ptr = (struct option_t*) p;
  char* init = NULL;

  if (!ptr)
    return;
  if (FuncTable[DTYPE (ptr->type)].opt_fromstr) {
    init_expand (&init, ptr);
    if (!FuncTable[DTYPE (ptr->type)].opt_fromstr (ptr, init, errbuf,
                                                       sizeof(errbuf))) {
      if (!option (OPTNOCURSES))
        mutt_endwin (NULL);
      fprintf (stderr, _("Invalid default setting for $%s found: \"%s\".\n"
                         "Please report this error: \"%s\"\n"),
               ptr->option, NONULL (init), errbuf);
      exit (1);
    }
    p_delete(&init);
  }

  set_option (OPTFORCEREDRAWINDEX);
  set_option (OPTFORCEREDRAWPAGER);
  set_option (OPTSORTSUBTHREADS);
  set_option (OPTNEEDRESORT);
  set_option (OPTRESORTINIT);
  set_option (OPTREDRAWTREE);
}

static int check_num (const char* option, unsigned long p,
                      char* errbuf, ssize_t errlen) {
  if ((int) p < 0) {
    if (errbuf)
      snprintf (errbuf, errlen, _("'%d' is invalid for $%s"), (int) p, option);
    return 0;
  }
  return 1;
}

static int check_history (const char* option __attribute__ ((unused)), unsigned long p,
                          char* errbuf, ssize_t errlen) {
  if (!check_num ("history", p, errbuf, errlen))
    return 0;
  mutt_init_history ();
  return 1;
}

static int check_special (const char* name, unsigned long val,
                          char* errbuf, ssize_t errlen) {
  int i = 0;

  for (i = 0; SpecialVars[i].name; i++) {
    if (m_strcmp(SpecialVars[i].name, name) == 0) {
      return (SpecialVars[i].check (SpecialVars[i].name,
                                    val, errbuf, errlen));
    }
  }
  return 1;
}

static const struct mapping_t* get_sortmap (struct option_t* option) {
  const struct mapping_t* map = NULL;

  switch (option->type & DT_SUBTYPE_MASK) {
  case DT_SORT_ALIAS:
    map = SortAliasMethods;
    break;
  case DT_SORT_BROWSER:
    map = SortBrowserMethods;
    break;
  case DT_SORT_KEYS:
    map = SortKeyMethods;
    break;
  case DT_SORT_AUX:
    map = SortAuxMethods;
    break;
  default:
    map = SortMethods;
    break;
  }
  return map;
}

static int parse_set (BUFFER * tmp, BUFFER * s, unsigned long data,
                      BUFFER * err)
{
  int query, unset, inv, reset, r = 0;
  struct option_t* option = NULL;

  while (MoreArgs (s)) {
    /* reset state variables */
    query = 0;
    unset = data & M_SET_UNSET;
    inv = data & M_SET_INV;
    reset = data & M_SET_RESET;

    if (*s->dptr == '?') {
      query = 1;
      s->dptr++;
    }
    else if (m_strncmp("no", s->dptr, 2) == 0) {
      s->dptr += 2;
      unset = !unset;
    }
    else if (m_strncmp("inv", s->dptr, 3) == 0) {
      s->dptr += 3;
      inv = !inv;
    }
    else if (*s->dptr == '&') {
      reset = 1;
      s->dptr++;
    }

    /* get the variable name */
    mutt_extract_token (tmp, s, M_TOKEN_EQUAL);
    option = hash_find(ConfigOptions, tmp->data);
    if (!option && !(reset && m_strcmp("all", tmp->data) == 0)) {
      snprintf (err->data, err->dsize, _("%s: unknown variable"), tmp->data);
      return -1;
    }
    s->dptr = vskipspaces(s->dptr);

    if (reset) {
      if (query || unset || inv) {
        snprintf (err->data, err->dsize, _("prefix is illegal with reset"));
        return -1;
      }

      if (s && *s->dptr == '=') {
        snprintf (err->data, err->dsize, _("value is illegal with reset"));
        return -1;
      }

      if (!m_strcmp("all", tmp->data)) {
        if (CurrentMenu == MENU_PAGER) {
          snprintf (err->data, err->dsize, _("Not available in this menu."));
          return -1;
        }
        hash_map (ConfigOptions, mutt_restore_default, 1);
        set_option (OPTFORCEREDRAWINDEX);
        set_option (OPTFORCEREDRAWPAGER);
        set_option (OPTSORTSUBTHREADS);
        set_option (OPTNEEDRESORT);
        set_option (OPTRESORTINIT);
        set_option (OPTREDRAWTREE);
        return 0;
      } else {
        mutt_restore_default (NULL, option, 1);
      }
    }
    else if (DTYPE (option->type) == DT_BOOL) {
      /* XXX this currently ignores the function table
       * as we don't get invert and stuff into it */
      if (s && *s->dptr == '=') {
        if (unset || inv || query) {
          snprintf (err->data, err->dsize, "Usage: set variable=yes|no");
          return -1;
        }

        s->dptr++;
        mutt_extract_token (tmp, s, 0);
        if (ascii_strcasecmp ("yes", tmp->data) == 0)
          unset = inv = 0;
        else if (ascii_strcasecmp ("no", tmp->data) == 0)
          unset = 1;
        else {
          snprintf (err->data, err->dsize, "Usage: set variable=yes|no");
          return -1;
        }
      }

      if (query) {
        bool_to_string (err->data, err->dsize, option);
        return 0;
      }

      if (unset)
        unset_option (option->data);
      else if (inv)
        toggle_option (option->data);
      else
        set_option (option->data);
    }
    else if (DTYPE (option->type) == DT_STR ||
             DTYPE (option->type) == DT_PATH ||
             DTYPE (option->type) == DT_NUM ||
             DTYPE (option->type) == DT_SORT ||
             DTYPE (option->type) == DT_RX)
    {
      /* XXX maybe we need to get unset into handlers? */
      if (DTYPE (option->type) == DT_STR || DTYPE (option->type) == DT_PATH) {
        if (unset) {
          p_delete((void **)(void *)&option->data);
          break;
        }
      }

      if (query || *s->dptr != '=') {
        FuncTable[DTYPE (option->type)].opt_tostr
          (err->data, err->dsize, option);
        break;
      }

      s->dptr++;
      mutt_extract_token (tmp, s, 0);
      if (!FuncTable[DTYPE (option->type)].opt_fromstr
          (option, tmp->data, err->data, err->dsize))
        r = -1;
    }
    else if (DTYPE (option->type) == DT_QUAD) {

      if (query) {
        quad_to_string (err->data, err->dsize, option);
        break;
      }

      if (*s->dptr == '=') {
        s->dptr++;
        mutt_extract_token (tmp, s, 0);
        if (ascii_strcasecmp ("yes", tmp->data) == 0)
          set_quadoption (option->data, M_YES);
        else if (ascii_strcasecmp ("no", tmp->data) == 0)
          set_quadoption (option->data, M_NO);
        else if (ascii_strcasecmp ("ask-yes", tmp->data) == 0)
          set_quadoption (option->data, M_ASKYES);
        else if (ascii_strcasecmp ("ask-no", tmp->data) == 0)
          set_quadoption (option->data, M_ASKNO);
        else {
          snprintf (err->data, err->dsize, _("'%s' is invalid for $%s\n"),
                    tmp->data, option->option);
          r = -1;
          break;
        }
      }
      else {
        if (inv)
          toggle_quadoption (option->data);
        else if (unset)
          set_quadoption (option->data, M_NO);
        else
          set_quadoption (option->data, M_YES);
      }
    }
    else {
      snprintf (err->data, err->dsize, _("%s: unknown type"),
                option->option);
      r = -1;
      break;
    }

    set_option (OPTFORCEREDRAWINDEX);
    set_option (OPTFORCEREDRAWPAGER);
    set_option (OPTSORTSUBTHREADS);
    set_option (OPTNEEDRESORT);
    set_option (OPTRESORTINIT);
    set_option (OPTREDRAWTREE);
  }
  return r;
}

#define MAXERRS 128

/* reads the specified initialization file.  returns -1 if errors were found
   so that we can pause to let the user know...  */
static int source_rc (const char *rcfile, BUFFER * err)
{
  FILE *f;
  int line = 0, rc = 0, conv = 0;
  BUFFER token;
  char *linebuf = NULL;
  char *currentline = NULL;
  ssize_t buflen;
  pid_t pid;

  if ((f = mutt_open_read (rcfile, &pid)) == NULL) {
    snprintf (err->data, err->dsize, "%s: %s", rcfile, strerror (errno));
    return -1;
  }

  p_clear(&token, 1);
  while ((linebuf = mutt_read_line(linebuf, &buflen, f, &line)) != NULL) {
    conv = ConfigCharset && (*ConfigCharset) && mod_cset.charset;
    if (conv) {
      currentline = m_strdup(linebuf);
      if (!currentline)
        continue;
      mutt_convert_string (&currentline, ConfigCharset, mod_cset.charset, 0);
    }
    else
      currentline = linebuf;

    CurRCLine = line;
    CurRCFile = rcfile;

    if (mutt_parse_rc_line (currentline, &token, err) == -1) {
      mutt_error (_("Error in %s, line %d: %s"), rcfile, line, err->data);
      if (--rc < -MAXERRS) {
        if (conv)
          p_delete(&currentline);
        break;
      }
    }
    else {
      if (rc < 0)
        rc = -1;
    }
    if (conv)
      p_delete(&currentline);
  }
  p_delete(&token.data);
  p_delete(&linebuf);
  m_fclose(&f);
  if (pid != -1)
    mutt_wait_filter (pid);
  if (rc) {
    /* the muttrc source keyword */
    snprintf (err->data, err->dsize,
              rc >= -MAXERRS ? _("source: errors in %s")
              : _("source: reading aborted due too many errors in %s"),
              rcfile);
    rc = -1;
  }
  return rc;
}

#undef MAXERRS

static int parse_source (BUFFER * tmp, BUFFER * s,
                         unsigned long data __attribute__ ((unused)),
                         BUFFER * err)
{
  char path[_POSIX_PATH_MAX];
  int rc = 0;

  do {
    if (mutt_extract_token (tmp, s, 0) != 0) {
      snprintf (err->data, err->dsize, _("source: error at %s"), s->dptr);
      return -1;
    }

    m_strcpy(path, sizeof(path), tmp->data);
    mutt_expand_path (path, sizeof(path));

    rc += source_rc (path, err);
  }
  while (MoreArgs (s));

  return (rc < 0) ? -1 : 0;
}

/* line		command to execute

   token	scratch buffer to be used by parser.  caller should free
   		token->data when finished.  the reason for this variable is
		to avoid having to allocate and deallocate a lot of memory
		if we are parsing many lines.  the caller can pass in the
		memory to use, which avoids having to create new space for
		every call to this function.

   err		where to write error messages */
int mutt_parse_rc_line (const char *line, BUFFER * token, BUFFER * err)
{
  int i, r = -1;
  BUFFER expn;

  p_clear(&expn, 1);
  expn.data = expn.dptr = line;
  expn.dsize = m_strlen(line);

  *err->data = 0;

  expn.dptr = vskipspaces(expn.dptr);
  while (*expn.dptr) {
    if (*expn.dptr == '#')
      break;                    /* rest of line is a comment */
    if (*expn.dptr == ';') {
      expn.dptr++;
      continue;
    }
    mutt_extract_token (token, &expn, 0);
    for (i = 0; Commands[i].name; i++) {
      if (!m_strcmp(token->data, Commands[i].name)) {
        if (Commands[i].func (token, &expn, Commands[i].data, err) != 0)
          goto finish;
        break;
      }
    }
    if (!Commands[i].name) {
      snprintf (err->data, err->dsize, _("%s: unknown command"),
                NONULL (token->data));
      goto finish;
    }
  }
  r = 0;
finish:
  if (expn.destroy)
    p_delete(&expn.data);
  return r;
}


#define NUMVARS (sizeof(MuttVars)/sizeof(MuttVars[0]))
#define NUMCOMMANDS (sizeof(Commands)/sizeof(Commands[0]))
/* initial string that starts completion. No telling how much crap
 * the user has typed so far. Allocate LONG_STRING just to be sure! */
char User_typed[LONG_STRING] = { 0 };

int Num_matched = 0;            /* Number of matches for completion */
char Completed[STRING] = { 0 }; /* completed string (command or variable) */
const char *Matches[MAX (NUMVARS, NUMCOMMANDS) + 1];  /* all the matches + User_typed */

/* helper function for completion.  Changes the dest buffer if
   necessary/possible to aid completion.
	dest == completion result gets here.
	src == candidate for completion.
	try == user entered data for completion.
	len == length of dest buffer.
*/
static void candidate (char *dest, char *try, const char *src, int len)
{
  int l;

  if (strstr (src, try) == src) {
    Matches[Num_matched++] = src;
    if (dest[0] == 0)
      m_strcpy(dest, len, src);
    else {
      for (l = 0; src[l] && src[l] == dest[l]; l++);
      dest[l] = 0;
    }
  }
}

int mutt_command_complete (char *buffer, ssize_t len, int pos, int numtabs)
{
  char *pt = buffer;
  int num;
  int spaces;                   /* keep track of the number of leading spaces on the line */

  buffer = vskipspaces(buffer);
  spaces = buffer - pt;

  pt = buffer + pos - spaces;
  while ((pt > buffer) && !isspace ((unsigned char) *pt))
    pt--;

  if (pt == buffer) {           /* complete cmd */
    /* first TAB. Collect all the matches */
    if (numtabs == 1) {
      Num_matched = 0;
      m_strcpy(User_typed, sizeof(User_typed), pt);
      p_clear(Matches, countof(Matches));
      p_clear(Completed, countof(Completed));
      for (num = 0; Commands[num].name; num++)
        candidate (Completed, User_typed, Commands[num].name,
                   sizeof(Completed));
      Matches[Num_matched++] = User_typed;

      /* All matches are stored. Longest non-ambiguous string is ""
       * i.e. dont change 'buffer'. Fake successful return this time */
      if (User_typed[0] == 0)
        return 1;
    }

    if (Completed[0] == 0 && User_typed[0])
      return 0;

    /* Num_matched will _always_ be atleast 1 since the initial
     * user-typed string is always stored */
    if (numtabs == 1 && Num_matched == 2)
      snprintf (Completed, sizeof(Completed), "%s", Matches[0]);
    else if (numtabs > 1 && Num_matched > 2)
      /* cycle thru all the matches */
      snprintf (Completed, sizeof(Completed), "%s",
                Matches[(numtabs - 2) % Num_matched]);

    /* return the completed command */
    m_strcpy(buffer, len - spaces, Completed);
  }
  else if (!m_strncmp(buffer, "set", 3)
           || !m_strncmp(buffer, "unset", 5)
           || !m_strncmp(buffer, "reset", 5)
           || !m_strncmp(buffer, "toggle", 6)) {    /* complete variables */
    const char *prefixes[] = { "no", "inv", "?", "&", NULL };

    pt++;
    /* loop through all the possible prefixes (no, inv, ...) */
    if (!m_strncmp(buffer, "set", 3)) {
      for (num = 0; prefixes[num]; num++) {
        if (!m_strncmp(pt, prefixes[num], m_strlen(prefixes[num]))) {
          pt += m_strlen(prefixes[num]);
          break;
        }
      }
    }

    /* first TAB. Collect all the matches */
    if (numtabs == 1) {
      Num_matched = 0;
      m_strcpy(User_typed, sizeof(User_typed), pt);
      p_clear(Matches, countof(Matches));
      p_clear(Completed, countof(Completed));
      for (num = 0; MuttVars[num].option; num++)
        candidate(Completed, User_typed, MuttVars[num].option,
                  sizeof(Completed));
      Matches[Num_matched++] = User_typed;

      /* All matches are stored. Longest non-ambiguous string is ""
       * i.e. dont change 'buffer'. Fake successful return this time */
      if (User_typed[0] == 0)
        return 1;
    }

    if (Completed[0] == 0 && User_typed[0])
      return 0;

    /* Num_matched will _always_ be atleast 1 since the initial
     * user-typed string is always stored */
    if (numtabs == 1 && Num_matched == 2)
      snprintf (Completed, sizeof(Completed), "%s", Matches[0]);
    else if (numtabs > 1 && Num_matched > 2)
      /* cycle thru all the matches */
      snprintf (Completed, sizeof(Completed), "%s",
                Matches[(numtabs - 2) % Num_matched]);

    m_strcpy(pt, buffer + len - pt - spaces, Completed);
  }
  else if (!m_strncmp(buffer, "exec", 4)) {
    struct binding_t *menu = km_get_table (CurrentMenu);

    if (!menu && CurrentMenu != MENU_PAGER)
      menu = OpGeneric;

    pt++;
    /* first TAB. Collect all the matches */
    if (numtabs == 1) {
      Num_matched = 0;
      m_strcpy(User_typed, sizeof(User_typed), pt);
      p_clear(Matches, countof(Matches));
      p_clear(Completed, countof(Completed));
      for (num = 0; menu[num].name; num++)
        candidate (Completed, User_typed, menu[num].name, sizeof(Completed));
      /* try the generic menu */
      if (Completed[0] == 0 && CurrentMenu != MENU_PAGER) {
        menu = OpGeneric;
        for (num = 0; menu[num].name; num++)
          candidate (Completed, User_typed, menu[num].name,
                     sizeof(Completed));
      }
      Matches[Num_matched++] = User_typed;

      /* All matches are stored. Longest non-ambiguous string is ""
       * i.e. dont change 'buffer'. Fake successful return this time */
      if (User_typed[0] == 0)
        return 1;
    }

    if (Completed[0] == 0 && User_typed[0])
      return 0;

    /* Num_matched will _always_ be atleast 1 since the initial
     * user-typed string is always stored */
    if (numtabs == 1 && Num_matched == 2)
      snprintf (Completed, sizeof(Completed), "%s", Matches[0]);
    else if (numtabs > 1 && Num_matched > 2)
      /* cycle thru all the matches */
      snprintf (Completed, sizeof(Completed), "%s",
                Matches[(numtabs - 2) % Num_matched]);

    m_strcpy(pt, buffer + len - pt - spaces, Completed);
  }
  else
    return 0;

  return 1;
}

int mutt_var_value_complete (char *buffer, ssize_t len, int pos)
{
  char var[STRING], *pt = buffer;
  int spaces;
  struct option_t* option = NULL;

  if (buffer[0] == 0)
    return 0;

  buffer = vskipspaces(buffer);
  spaces = buffer - pt;

  pt = buffer + pos - spaces;
  while ((pt > buffer) && !isspace ((unsigned char) *pt))
    pt--;
  pt++;                         /* move past the space */
  if (*pt == '=')               /* abort if no var before the '=' */
    return 0;

  if (m_strncmp(buffer, "set", 3) == 0) {
    m_strcpy(var, sizeof(var), pt);
    /* ignore the trailing '=' when comparing */
    var[m_strlen(var) - 1] = 0;
    if (!(option = hash_find (ConfigOptions, var)))
      return 0;                 /* no such variable. */
    else {
      char tmp[LONG_STRING], tmp2[LONG_STRING];
      char *s, *d;
      ssize_t dlen = buffer + len - pt - spaces;
      const char *vals[] = { "no", "yes", "ask-no", "ask-yes" };

      tmp[0] = '\0';

      if ((DTYPE (option->type) == DT_STR) ||
          (DTYPE (option->type) == DT_PATH) ||
          (DTYPE (option->type) == DT_RX)) {
        m_strcpy(tmp, sizeof(tmp), NONULL(*((char **)option->data)));
        if (DTYPE (option->type) == DT_PATH)
          mutt_pretty_mailbox (tmp);
      }
      else if (DTYPE (option->type) == DT_QUAD)
        m_strcpy(tmp, sizeof(tmp), vals[quadoption(option->data)]);
      else if (DTYPE (option->type) == DT_NUM)
        snprintf (tmp, sizeof(tmp), "%d", (*((short *) option->data)));
      else if (DTYPE (option->type) == DT_SORT) {
        const struct mapping_t *map;
        const char *p;

        switch (option->type & DT_SUBTYPE_MASK) {
        case DT_SORT_ALIAS:
          map = SortAliasMethods;
          break;
        case DT_SORT_BROWSER:
          map = SortBrowserMethods;
          break;
        case DT_SORT_KEYS:
          map = SortKeyMethods;
          break;
        default:
          map = SortMethods;
          break;
        }
        p = mutt_getnamebyvalue(*((short *) option->data) & SORT_MASK, map);
        snprintf(tmp, sizeof(tmp), "%s%s%s",
                 (*((short *)option->data) & SORT_REVERSE) ? "reverse-" : "",
                 (*((short *)option->data) & SORT_LAST) ? "last-" : "", p);
      }
      else if (DTYPE (option->type) == DT_BOOL)
        m_strcpy(tmp, sizeof(tmp), option(option->data) ? "yes" : "no");
      else
        return 0;

      for (s = tmp, d = tmp2; *s && (d - tmp2) < ssizeof(tmp2) - 2;) {
        if (*s == '\\' || *s == '"')
          *d++ = '\\';
        *d++ = *s++;
      }
      *d = '\0';

      m_strcpy(tmp, sizeof(tmp), pt);
      snprintf (pt, dlen, "%s\"%s\"", tmp, tmp2);

      return 1;
    }
  }
  return 0;
}

static int mutt_execute_commands (string_list_t * p)
{
  BUFFER err, token;
  char errstr[STRING];

  p_clear(&err, 1);
  err.data = errstr;
  err.dsize = sizeof(errstr);
  p_clear(&token, 1);
  for (; p; p = p->next) {
    if (mutt_parse_rc_line (p->data, &token, &err) != 0) {
      fprintf (stderr, _("Error in command line: %s\n"), err.data);
      p_delete(&token.data);
      return -1;
    }
  }
  p_delete(&token.data);
  return 0;
}

void mutt_init (int skip_sys_rc, string_list_t * commands)
{
  struct passwd *pw;
  const char *p;
  char buffer[STRING], error[STRING];
  int default_rc = 0, need_pause = 0;
  int i;
  BUFFER err;

  p_clear(&err, 1);
  err.data = error;
  err.dsize = sizeof(error);

  ConfigOptions = hash_new (sizeof(MuttVars) * 2, 0);
  for (i = 0; MuttVars[i].option; i++) {
    hash_insert (ConfigOptions, MuttVars[i].option, &MuttVars[i]);
  }

  /*
   * XXX - use something even more difficult to predict?
   */
  snprintf (AttachmentMarker, sizeof(AttachmentMarker),
            "\033]9;%ld\a", (long) time (NULL));

  luaM_initialize();
  /* Get some information about the user */
  if ((pw = getpwuid (getuid ()))) {
    char rnbuf[STRING];
    mutt_gecos_name(rnbuf, sizeof(rnbuf), pw, mod_core.gecos_mask);
    Realname = m_strdup(rnbuf);
  }

  if ((p = getenv("MAIL") ?: getenv("MAILDIR"))) {
    Spoolfile = m_strdup(p);
  } else {
    mutt_concat_path(buffer, sizeof(buffer), NONULL(mod_core.homedir), MAILPATH);
    Spoolfile = m_strdup(buffer);
  }

  if ((p = getenv ("REPLYTO")) != NULL) {
    BUFFER buf, token;

    snprintf (buffer, sizeof(buffer), "Reply-To: %s", p);

    p_clear(&buf, 1);
    buf.data = buf.dptr = buffer;
    buf.dsize = m_strlen(buffer);

    p_clear(&token, 1);
    parse_my_hdr (&token, &buf, 0, &err);
    p_delete(&token.data);
  }

  /* Set standard defaults */
  hash_map (ConfigOptions, mutt_set_default, 0);
  hash_map (ConfigOptions, mutt_restore_default, 0);

  CurrentMenu = MENU_MAIN;
  mutt_init_history ();

  if (!Muttrc) {
      snprintf (buffer, sizeof(buffer), "%s/.madmuttrc", NONULL(mod_core.homedir));
    if (access (buffer, F_OK) == -1)
      snprintf (buffer, sizeof(buffer), "%s/.madmutt/madmuttrc",
                NONULL(mod_core.homedir));

    default_rc = 1;
    Muttrc = m_strdup(buffer);
  }
  else {
    m_strcpy(buffer, sizeof(buffer), Muttrc);
    p_delete(&Muttrc);
    mutt_expand_path (buffer, sizeof(buffer));
    Muttrc = m_strdup(buffer);
  }

  /* Process the global rc file if it exists and the user hasn't explicity
     requested not to via "-n".  */
  if (!skip_sys_rc) {
    snprintf (buffer, sizeof(buffer), "%s/Madmuttrc-%s", SYSCONFDIR,
              MUTT_VERSION);
    if (access (buffer, F_OK) == -1)
      snprintf (buffer, sizeof(buffer), "%s/Madmuttrc", SYSCONFDIR);
    if (access (buffer, F_OK) == -1)
      snprintf (buffer, sizeof(buffer), "%s/Madmuttrc-%s", PKGDATADIR,
                MUTT_VERSION);
    if (access (buffer, F_OK) == -1)
      snprintf (buffer, sizeof(buffer), "%s/Madmuttrc", PKGDATADIR);
    if (access (buffer, F_OK) != -1) {
      if (source_rc (buffer, &err) != 0) {
        fputs (err.data, stderr);
        fputc ('\n', stderr);
        need_pause = 1;
      }
    }
  }

  /* Read the user's initialization file.  */
  if (access (Muttrc, F_OK) != -1) {
    if (!option (OPTNOCURSES))
      mutt_endwin (NULL);
    if (source_rc (Muttrc, &err) != 0) {
      fputs (err.data, stderr);
      fputc ('\n', stderr);
      need_pause = 1;
    }
  }
  else if (!default_rc) {
    /* file specified by -F does not exist */
    snprintf (buffer, sizeof(buffer), "%s: %s", Muttrc, strerror (errno));
    mutt_endwin (buffer);
    exit (1);
  }

  /* LUA {{{ */
  snprintf(buffer, sizeof(buffer), "%s/.madmutt.lua", NONULL(mod_core.homedir));
  if (access(buffer, F_OK) < 0)
      snprintf(buffer, sizeof(buffer), "%s/.madmutt/cfg.lua", NONULL(mod_core.homedir));
  if (!access(buffer, F_OK)) {
      need_pause = luaM_wrap(mutt_error, luaM_dofile(buffer));
  }
  /* }}} */

  if (mutt_execute_commands (commands) != 0)
    need_pause = 1;

  /* warn about synonym variables */
  if (Synonyms) {
    syn_t *syn;

    fprintf (stderr, _("Warning: the following synonym variables were found:\n"));

    for (syn = Synonyms; syn; syn = syn->next) {
      fprintf(stderr, "$%s ($%s should be used) (%s:%d)\n",
              syn->o ? NONULL(syn->o->option) : "",
              syn->n ? NONULL(syn->n->option) : "",
              NONULL(syn->f), syn->l);
    }
    fprintf (stderr, _("Warning: synonym variables are scheduled"
                       " for removal.\n"));
    syn_list_wipe(&Synonyms);
    need_pause = 1;
  }

  if (need_pause && !option (OPTNOCURSES)) {
    if (mutt_any_key_to_continue (NULL) == -1)
      mutt_exit (1);
  }
}

